/*
 * $Id: Browser.java,v 1.3 1998/10/22 23:03:15 begolej Exp $
 *
 * Copyright (c) 1998 James Begole,  All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies.
 *
 * James Begole MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE
 * SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING
 * BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT.
 * James Begole SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY
 * LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS
 * SOFTWARE OR ITS DERIVATIVES.
 */

import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import sunw.hotjava.bean.*;

/**
 * An HTML browser composed of HotJava HTML components.
 * The currently viewed URL is a "bound property," which
 * notifies registered listeners when there is a change.
 * A collaborative environment will register as a listener,
 * distribute the property change to replicas of the browser,
 * which will change their currently viewed URL to match.
 *
 * @version     $Revision: 1.3 $
 * @author  James "Bo" Begole
 */
public class Browser extends Panel {

  // Create the browser state, stack, and authenticator beans
  /** Maintains browser state. Required by HotJava Bean */
  HotJavaSystemState sysState = new HotJavaSystemState();

  /** Maintains stack of viewed URLs. Required by HotJava Bean */
  HotJavaDocumentStack docStack = new HotJavaDocumentStack();

  /** Browser security. Required by HotJava Bean */
  AuthenticatorBean authenticator = new AuthenticatorBean();

  /** A string representation of the currently viewed URL */
  String docString = null;

  /** A text input field */
  TextBean textBean = null;

  /** The browser itself.
      Displays the URL indicated by docString
  */
  HotJavaBrowserBean browserBean = new HotJavaBrowserBean();

  /** Maintains list of of PropertyChangeListeners.
  Notifies PropertyChangeListeners when a bound property
  is modified.
  */
  PropertyChangeSupport changes = new PropertyChangeSupport(this);

  /** Construct a simple browser using the HotJava bean
  components.
  */
  public Browser() {
    setLayout(new BorderLayout());

    // create panel to contain input text field, back,
    // and forward buttons
    Panel controlPanel = new Panel();
    controlPanel.setLayout(new FlowLayout(FlowLayout.CENTER, 10, 5));
    Button back = (Button)controlPanel.add(new Button("Back"));
    Button forward = (Button)controlPanel.add(new Button("Forward"));
    textBean = (TextBean)controlPanel.add(new TextBean());

    // listen for changes to text
    textBean.addPropertyChangeListener(new PropertyChangeListener() {
      public void propertyChange(PropertyChangeEvent evt) {
        String prop = evt.getPropertyName();
        if ("text".equals(prop)) {
          /*
           * when the text changesin textBean,
           * reset the docString and load the new URL
           */
          setDocumentString((String)evt.getNewValue());
        }
      }
    });

    // set the interface controls at the top of the frame
    add("North", controlPanel);

    // and put the displayed URL in the center
    add("Center", browserBean);

    // listen for changes to the current URL
    browserBean.addPropertyChangeListener(
                                  new PropertyChangeListener() {
      public void propertyChange(PropertyChangeEvent evt) {
        String prop = evt.getPropertyName();
        if ("documentString".equals(prop)) {
          /*
           * if the documentString property was changed (perhaps
           * by a remote collaborator), reset the docString, load
           * the new URL and display the new string in textBean
           */
          setDocumentString((String)evt.getNewValue());
        }
      }
    });

    /* establish listener connections among the browser state,
     * stack, and authenticator beans
     */
    BrowserHL bHL = new BrowserHL();  // history list (i.e., stack)
    bHL.setTarget(browserBean);
    docStack.addBrowserHistoryListener(bHL);
    StackHL sHL = new StackHL();
    sHL.setTarget(docStack);
    browserBean.addBrowserHistoryListener(sHL);

    /* Define the actions to take when user presses "back" and
     * "forward" buttons.
     */
    BackActionListener bAL = new BackActionListener();
    bAL.setTarget(docStack);
    back.addActionListener(bAL);
    ForwardActionListener fAL = new ForwardActionListener();
    fAL.setTarget(docStack);
    forward.addActionListener(fAL);
  }

  /**
   * Get the preferred size for this component.
   */
  public Dimension getPreferredSize() {
    Dimension currentSize = getSize();
    Dimension minSize = getMinimumSize();
    if (lessThan(currentSize, minSize))
      return minSize;
    else
      return currentSize;
  }

  /**
   * Utility to determin if one dimension is "smaller"
   * than another in width or height (not necessarily total area)
   * @param lhs left-hand-side of comparison (lhs < rhs)
   * @param rhs right-hand-side of comparison (lhs < rhs)
   */
  public static boolean lessThan(Dimension lhs, Dimension rhs) {
    if (lhs.width < rhs.width)
      return true;
    else
      return (lhs.height < rhs.height);
  }

  /** minimum size for this component */
  Dimension minSize = new Dimension(400, 500);

  /**
   * Get the minimum size for this component.
   */
  public Dimension getMinimumSize() {
    return minSize;
  }

  /**
   * Set the string that represents the URL currently viewed by
   * the browser.
   */
  public void setDocumentString(String newString) {
    // only change the current docString, if newString is different
    if (!(newString.equals(docString))) {
      // display the new docString in textBean
      textBean.setText(newString);
      // load the new docString into the browser
      browserBean.setDocumentString(newString);
      /*
       * Notify all listeners that the docString has changed.
       * This is where JavaBeans provides support for collaboration.
       * A listener will distribute the property change to
       * replicas of the browser, which will change their docString
       * to match.
       */
      changes.firePropertyChange("documentString", docString,
                                 newString);
      docString = newString;
    }
  }

  /**
   * Get the string the represents the currently loaded URL.
   */
   public String getDocumentString() {
     return docString;
   }

   /**
    * Allows PropertyChangeListeners to register with the browser.
    * The collaborative environment will register as a listener.
    *
    * @param l The PropertyChangeListener to be added
    */
    public void addPropertyChangeListener(PropertyChangeListener l) {
        changes.addPropertyChangeListener(l);
    }
    /**
     * Remove a PropertyChangeListener
     *
     * @param l The PropertyChangeListener to be removed
     */
  public void removePropertyChangeListener(PropertyChangeListener l){
    changes.removePropertyChangeListener(l);
  }
}

/**
  The following classes specify the actions to take to display
  URLs from the browser's stack.
*/
class BrowserHL implements BrowserHistoryListener {
    public void setTarget(HotJavaBrowserBean t) {
        target = t;
    }
    public void executeHistoryCommand(BrowserHistoryEvent arg0) {
       target.executeHistoryCommand(arg0);
    }
    private HotJavaBrowserBean target;
}

class StackHL implements BrowserHistoryListener {
    public void setTarget(HotJavaDocumentStack t) {
        target = t;
    }
    public void executeHistoryCommand(BrowserHistoryEvent arg0) {
        target.executeHistoryCommand(arg0);
    }
    private HotJavaDocumentStack target;
}

/**
 * Specifies the action to take when the "back" button is pressed.
 */
class BackActionListener implements ActionListener {
    public void setTarget(HotJavaDocumentStack t) {
        target = t;
    }
    public void actionPerformed(ActionEvent arg0) {
        target.previousDocument();
    }
    private HotJavaDocumentStack target;
}

/**
 * Specifies the action to take when the "forward" button is pressed.
 */
class ForwardActionListener implements ActionListener {
    public void setTarget(HotJavaDocumentStack t) {
        target = t;
    }
    public void actionPerformed(ActionEvent arg0) {
        target.nextDocument();
    }
    private HotJavaDocumentStack target;
}
